import os
import threading
import time as ttime
import warnings

import numpy as np
from bec_lib import MessageEndpoints, bec_logger, messages
from ophyd import Component as Cpt
from ophyd import DynamicDeviceComponent as Dcpt
from ophyd import Device, DeviceStatus, Kind
from ophyd import PositionerBase, Signal
from ophyd.sim import SynSignal

from ophyd.utils import LimitError
from ophyd_devices.utils.bec_scaninfo_mixin import BecScaninfoMixin
from ophyd_devices.sim.sim_data import SimulatedDataBase, SimulatedDataCamera, SimulatedDataMonitor
from ophyd_devices.sim.sim_test_devices import DummyController

from ophyd_devices.sim.sim_signals import SetableSignal, ReadOnlySignal, ComputedReadOnlySignal

logger = bec_logger.logger


class DeviceStop(Exception):
    pass


class SimMonitor(Device):
    """
    A simulated device mimic any 1D Axis (position, temperature, beam).

    It's readback is a computed signal, which is configurable by the user and from the command line.
    The corresponding simulation class is sim_cls=SimulatedDataMonitor, more details on defaults within the simulation class.

    >>> monitor = SimMonitor(name="monitor")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argmuent, passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation, check simulation type defaults for more details.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.

    """

    USER_ACCESS = ["sim"]

    sim_cls = SimulatedDataMonitor

    readback = Cpt(ComputedReadOnlySignal, value=0, kind=Kind.hinted)

    SUB_READBACK = "readback"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        name,
        *,
        precision: int = 3,
        sim_init: dict = None,
        parent=None,
        kind=None,
        device_manager=None,
        **kwargs,
    ):
        self.precision = precision
        self.init_sim_params = sim_init
        self.sim = self.sim_cls(parent=self, device_manager=device_manager, **kwargs)

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self.sim.sim_state[self.name] = self.sim.sim_state.pop(self.readback.name, None)
        self.readback.name = self.name


class SimCamera(Device):
    """A simulated device mimic any 2D camera.

    It's image is a computed signal, which is configurable by the user and from the command line.
    The corresponding simulation class is sim_cls=SimulatedDataCamera, more details on defaults within the simulation class.

    >>> camera = SimCamera(name="camera")

    Parameters
    ----------
    name (string)           : Name of the device. This is the only required argmuent, passed on to all signals of the device.
    precision (integer)     : Precision of the readback in digits, written to .describe(). Default is 3 digits.
    sim_init (dict)         : Dictionary to initiate parameters of the simulation, check simulation type defaults for more details.
    parent                  : Parent device, optional, is used internally if this signal/device is part of a larger device.
    kind                    : A member the Kind IntEnum (or equivalent integer), optional. Default is Kind.normal. See Kind for options.
    device_manager          : DeviceManager from BEC, optional . Within startup of simulation, device_manager is passed on automatically.

    """

    USER_ACCESS = ["sim"]

    sim_cls = SimulatedDataCamera
    SHAPE = (100, 100)

    SUB_MONITOR = "monitor"
    _default_sub = SUB_MONITOR

    exp_time = Cpt(SetableSignal, name="exp_time", value=1, kind=Kind.config)
    file_path = Cpt(SetableSignal, name="file_path", value="", kind=Kind.config)
    file_pattern = Cpt(SetableSignal, name="file_pattern", value="", kind=Kind.config)
    frames = Cpt(SetableSignal, name="frames", value=1, kind=Kind.config)
    burst = Cpt(SetableSignal, name="burst", value=1, kind=Kind.config)

    image_shape = Cpt(SetableSignal, name="image_shape", value=SHAPE, kind=Kind.config)
    image = Cpt(
        ComputedReadOnlySignal,
        name="image",
        value=np.empty(SHAPE, dtype=np.uint16),
        kind=Kind.omitted,
    )

    def __init__(
        self,
        name,
        *,
        kind=None,
        parent=None,
        sim_init: dict = None,
        device_manager=None,
        **kwargs,
    ):
        self.device_manager = device_manager
        self.init_sim_params = sim_init
        self.sim = self.sim_cls(parent=self, device_manager=device_manager, **kwargs)

        super().__init__(name=name, parent=parent, kind=kind, **kwargs)
        self._stopped = False
        self._staged = False
        self.scaninfo = None
        self._update_scaninfo()

    def trigger(self) -> DeviceStatus:
        """Trigger the camera to acquire images.

        This method can be called from BEC during a scan. It will acquire images and send them to BEC.
        Whether the trigger is send from BEC is determined by the softwareTrigger argument in the device config.

        Here, we also run a callback on SUB_MONITOR to send the image data the device_monitor endpoint in BEC.
        """
        status = DeviceStatus(self)

        self.subscribe(status._finished, event_type=self.SUB_ACQ_DONE, run=False)

        def acquire():
            try:
                for _ in range(self.burst.get()):
                    self._run_subs(sub_type=self.SUB_MONITOR, value=self.image.get())
                    if self._stopped:
                        raise DeviceStop
            except DeviceStop:
                pass
            finally:
                self._stopped = False
                self._done_acquiring()

        threading.Thread(target=acquire, daemon=True).start()
        return status

    def _update_scaninfo(self) -> None:
        """Update scaninfo from BecScaninfoMixing
        This depends on device manager and operation/sim_mode
        """
        self.scaninfo = BecScaninfoMixin(self.device_manager)

    def stage(self) -> list[object]:
        """Stage the camera for upcoming scan

        This method is called from BEC in preparation of a scan.
        It receives metadata about the scan from BEC,
        compiles it and prepares the camera for the scan.

        FYI: No data is written to disk in the simulation, but upon each trigger it
        is published to the device_monitor endpoint in REDIS.
        """
        if self._staged:
            return super().stage()
        self.scaninfo.load_scan_metadata()
        self.file_path.set(
            os.path.join(
                self.file_path.get(), self.file_pattern.get().format(self.scaninfo.scan_number)
            )
        )
        self.frames.set(self.scaninfo.num_points * self.scaninfo.frames_per_trigger)
        self.exp_time.set(self.scaninfo.exp_time)
        self.burst.set(self.scaninfo.frames_per_trigger)
        self._stopped = False
        return super().stage()

    def unstage(self) -> list[object]:
        """Unstage the device

        Send reads from all config signals to redis
        """
        if self._stopped is True or not self._staged:
            return super().unstage()

        return super().unstage()

    def stop(self, *, success=False):
        """Stop the device"""
        self._stopped = True
        super().stop(success=success)


class SimPositioner(Device, PositionerBase):
    """
    A simulated device mimicing any 1D Axis device (position, temperature, rotation).

    Parameters
    ----------
    name : string, keyword only
    readback_func : callable, optional
        When the Device is set to ``x``, its readback will be updated to
        ``f(x)``. This can be used to introduce random noise or a systematic
        offset.
        Expected signature: ``f(x) -> value``.
    value : object, optional
        The initial value. Default is 0.
    delay : number, optional
        Simulates how long it takes the device to "move". Default is 0 seconds.
    precision : integer, optional
        Digits of precision. Default is 3.
    parent : Device, optional
        Used internally if this Signal is made part of a larger Device.
    kind : a member the Kind IntEnum (or equivalent integer), optional
        Default is Kind.normal. See Kind for options.
    """

    # Specify which attributes are accessible via BEC client
    USER_ACCESS = ["sim", "readback", "speed", "dummy_controller"]

    sim_cls = SimulatedDataBase

    # Define the signals as class attributes
    readback = Cpt(ReadOnlySignal, name="readback", value=0, kind=Kind.hinted)
    setpoint = Cpt(SetableSignal, value=0, kind=Kind.normal)
    motor_is_moving = Cpt(SetableSignal, value=0, kind=Kind.normal)

    # Config signals
    velocity = Cpt(SetableSignal, value=1, kind=Kind.config)
    acceleration = Cpt(SetableSignal, value=1, kind=Kind.config)

    # Ommitted signals
    high_limit_travel = Cpt(SetableSignal, value=0, kind=Kind.omitted)
    low_limit_travel = Cpt(SetableSignal, value=0, kind=Kind.omitted)
    unused = Cpt(Signal, value=1, kind=Kind.omitted)

    # TODO add short description to these two lines and explain what this does
    SUB_READBACK = "readback"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        *,
        name,
        readback_func=None,
        value=0,
        delay=1,
        speed=1,
        update_frequency=2,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        limits=None,
        tolerance: float = 0.5,
        sim: dict = None,
        **kwargs,
    ):
        # Whether motions should be instantaneous or depend on motor velocity
        self.delay = delay
        self.precision = precision
        self.tolerance = tolerance
        self.init_sim_params = sim

        self.speed = speed
        self.update_frequency = update_frequency
        self._stopped = False
        self.dummy_controller = DummyController()

        # initialize inner dictionary with simulated state
        self.sim = self.sim_cls(parent=self, **kwargs)

        super().__init__(name=name, labels=labels, kind=kind, **kwargs)
        # Rename self.readback.name to self.name, also in self.sim_state
        self.sim.sim_state[self.name] = self.sim.sim_state.pop(self.readback.name, None)
        self.readback.name = self.name
        # Init limits from deviceConfig
        if limits is not None:
            assert len(limits) == 2
            self.low_limit_travel.put(limits[0])
            self.high_limit_travel.put(limits[1])

    @property
    def limits(self):
        """Return the limits of the simulated device."""
        return (self.low_limit_travel.get(), self.high_limit_travel.get())

    @property
    def low_limit(self):
        """Return the low limit of the simulated device."""
        return self.limits[0]

    @property
    def high_limit(self):
        """Return the high limit of the simulated device."""
        return self.limits[1]

    def check_value(self, value: any):
        """
        Check that requested position is within existing limits.

        This function has to be implemented on the top level of the positioner.
        """
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not low_limit <= value <= high_limit:
            raise LimitError(f"position={value} not within limits {self.limits}")

    def _set_sim_state(self, signal_name: str, value: any) -> None:
        """Update the simulated state of the device."""
        self.sim.sim_state[signal_name]["value"] = value
        self.sim.sim_state[signal_name]["timestamp"] = ttime.time()

    def _get_sim_state(self, signal_name: str) -> any:
        """Return the simulated state of the device."""
        return self.sim.sim_state[signal_name]["value"]

    def move(self, value: float, **kwargs) -> DeviceStatus:
        """Change the setpoint of the simulated device, and simultaneously initiated a motion."""
        self._stopped = False
        self.check_value(value)
        old_setpoint = self._get_sim_state(self.setpoint.name)
        self._set_sim_state(self.motor_is_moving.name, 1)
        self._set_sim_state(self.setpoint.name, value)

        def update_state(val):
            """Update the state of the simulated device."""
            if self._stopped:
                raise DeviceStop
            old_readback = self._get_sim_state(self.readback.name)
            self._set_sim_state(self.readback.name, val)

            # Run subscription on "readback"
            self._run_subs(
                sub_type=self.SUB_READBACK,
                old_value=old_readback,
                value=self.sim.sim_state[self.readback.name]["value"],
                timestamp=self.sim.sim_state[self.readback.name]["timestamp"],
            )

        st = DeviceStatus(device=self)
        if self.delay:
            # If self.delay is not 0, we use the speed and updated frequency of the device to compute the motion
            def move_and_finish():
                """Move the simulated device and finish the motion."""
                success = True
                try:
                    # Compute final position with some jitter
                    move_val = self._get_sim_state(
                        self.setpoint.name
                    ) + self.tolerance * np.random.uniform(-1, 1)
                    # Compute the number of updates needed to reach the final position with the given speed
                    updates = np.ceil(
                        np.abs(old_setpoint - move_val) / self.speed * self.update_frequency
                    )
                    # Loop over the updates and update the state of the simulated device
                    for ii in np.linspace(old_setpoint, move_val, int(updates)):
                        ttime.sleep(1 / self.update_frequency)
                        update_state(ii)
                    # Update the state of the simulated device to the final position
                    update_state(move_val)
                    self._set_sim_state(self.motor_is_moving, 0)
                except DeviceStop:
                    success = False
                finally:
                    self._stopped = False
                # Call function from positioner base to indicate that motion finished with success
                self._done_moving(success=success)
                # Set status to finished
                st.set_finished()

            # Start motion in Thread
            threading.Thread(target=move_and_finish, daemon=True).start()

        else:
            # If self.delay is 0, we move the simulated device instantaneously
            update_state(value)
            self._done_moving()
            st.set_finished()
        return st

    def stop(self, *, success=False):
        """Stop the motion of the simulated device."""
        super().stop(success=success)
        self._stopped = True

    @property
    def position(self):
        """Return the current position of the simulated device."""
        return self.readback.get()

    @property
    def egu(self):
        """Return the engineering units of the simulated device."""
        return "mm"


class SynFlyer(Device, PositionerBase):
    def __init__(
        self,
        *,
        name,
        readback_func=None,
        value=0,
        delay=0,
        speed=1,
        update_frequency=2,
        precision=3,
        parent=None,
        labels=None,
        kind=None,
        device_manager=None,
        **kwargs,
    ):
        if readback_func is None:

            def readback_func(x):
                return x

        sentinel = object()
        loop = kwargs.pop("loop", sentinel)
        if loop is not sentinel:
            warnings.warn(
                f"{self.__class__} no longer takes a loop as input.  "
                "Your input will be ignored and may raise in the future",
                stacklevel=2,
            )
        self.sim_state = {}
        self._readback_func = readback_func
        self.delay = delay
        self.precision = precision
        self.tolerance = kwargs.pop("tolerance", 0.5)
        self.device_manager = device_manager

        # initialize values
        self.sim_state["readback"] = readback_func(value)
        self.sim_state["readback_ts"] = ttime.time()

        super().__init__(name=name, parent=parent, labels=labels, kind=kind, **kwargs)

    @property
    def hints(self):
        return {"fields": ["flyer_samx", "flyer_samy"]}

    def kickoff(self, metadata, num_pos, positions, exp_time: float = 0):
        positions = np.asarray(positions)

        def produce_data(device, metadata):
            buffer_time = 0.2
            elapsed_time = 0
            bundle = messages.BundleMessage()
            for ii in range(num_pos):
                bundle.append(
                    messages.DeviceMessage(
                        signals={
                            self.name: {
                                "flyer_samx": {"value": positions[ii, 0], "timestamp": 0},
                                "flyer_samy": {"value": positions[ii, 1], "timestamp": 0},
                            }
                        },
                        metadata={"pointID": ii, **metadata},
                    ).dumps()
                )
                ttime.sleep(exp_time)
                elapsed_time += exp_time
                if elapsed_time > buffer_time:
                    elapsed_time = 0
                    device.device_manager.producer.send(
                        MessageEndpoints.device_read(device.name), bundle.dumps()
                    )
                    bundle = messages.BundleMessage()
                    device.device_manager.producer.set_and_publish(
                        MessageEndpoints.device_status(device.name),
                        messages.DeviceStatusMessage(
                            device=device.name,
                            status=1,
                            metadata={"pointID": ii, **metadata},
                        ).dumps(),
                    )
            device.device_manager.producer.send(
                MessageEndpoints.device_read(device.name), bundle.dumps()
            )
            device.device_manager.producer.set_and_publish(
                MessageEndpoints.device_status(device.name),
                messages.DeviceStatusMessage(
                    device=device.name,
                    status=0,
                    metadata={"pointID": num_pos, **metadata},
                ).dumps(),
            )
            print("done")

        flyer = threading.Thread(target=produce_data, args=(self, metadata))
        flyer.start()


class SynDynamicComponents(Device):
    messages = Dcpt({f"message{i}": (SynSignal, None, {"name": f"msg{i}"}) for i in range(1, 6)})


class SynDeviceSubOPAAS(Device):
    zsub = Cpt(SimPositioner, name="zsub")


class SynDeviceOPAAS(Device):
    x = Cpt(SimPositioner, name="x")
    y = Cpt(SimPositioner, name="y")
    z = Cpt(SimPositioner, name="z")



import os
import time

from bec_lib.bec_service import SERVICE_CONFIG
from bec_lib.device import DeviceStatus
from bec_lib.file_utils import FileWriterMixin
from ophyd import Device
from ophyd.device import Staged
from ophyd_devices.utils.bec_scaninfo_mixin import BecScaninfoMixin
from ophyd_devices.utils import bec_utils


class DetectorInitError(Exception):
    """Raised when initiation of the device class fails,
    due to missing device manager or not started in sim_mode."""


class CustomDetectorMixin:
    """
    Mixin class for custom detector logic

    This class is used to implement BL specific logic for the detector.
    It is used in the PSIDetectorBase class.

    For the integration of a new detector, the following functions should
    help with integrating functionality, but additional ones can be added.

    Check PSIDetectorBase for the functions that are called during relevant function calls of
    stage, unstage, trigger, stop and _init.
    """

    def __init__(self, *_args, parent: Device = None, **_kwargs) -> None:
        self.parent = parent

    def initialize_default_parameter(self) -> None:
        """
        Init parameters for the detector

        Raises (optional):
            DetectorTimeoutError: if detector cannot be initialized
        """

    def initialize_detector(self) -> None:
        """
        Init parameters for the detector

        Raises (optional):
            DetectorTimeoutError: if detector cannot be initialized
        """

    def initialize_detector_backend(self) -> None:
        """
        Init parameters for teh detector backend (filewriter)

        Raises (optional):
            DetectorTimeoutError: if filewriter cannot be initialized
        """

    def prepare_detector(self) -> None:
        """
        Prepare detector for the scan
        """

    def prepare_detector_backend(self) -> None:
        """
        Prepare detector backend for the scan
        """

    def stop_detector(self) -> None:
        """
        Stop the detector
        """

    def stop_detector_backend(self) -> None:
        """
        Stop the detector backend
        """

    def on_trigger(self) -> None:
        """
        Specify actions to be executed upon receiving trigger signal
        """

    def pre_scan(self) -> None:
        """
        Specify actions to be executed right before a scan

        BEC calls pre_scan just before execution of the scan core.
        It is convenient to execute time critical features of the detector,
        e.g. arming it, but it is recommended to keep this function as short/fast as possible.
        """

    def finished(self) -> None:
        """
        Specify actions to be executed during unstage

        This may include checks if acquisition was succesful

        Raises (optional):
            DetectorTimeoutError: if detector cannot be stopped
        """

    def check_scanID(self) -> None:
        """
        Check if BEC is running on a new scanID
        """

    def publish_file_location(self, done: bool = False, successful: bool = None) -> None:
        """
        Publish the designated filepath from data backend to REDIS.

        Typically, the following two message types are published:

        - file_event: event for the filewriter
        - public_file: event for any secondary service (e.g. radial integ code)
        """

    def wait_for_signals(
        self,
        signal_conditions: list,
        timeout: float,
        check_stopped: bool = False,
        interval: float = 0.05,
        all_signals: bool = False,
    ) -> bool:
        """Wait for signals to reach a certain condition

        Args:
            signal_conditions (tuple): tuple of (get_current_state, condition) functions
            timeout (float): timeout in seconds
            interval (float): interval in seconds
            all_signals (bool): True if all signals should be True, False if any signal should be True
        Returns:
            bool: True if all signals are in the desired state, False if timeout is reached
        """
        timer = 0
        while True:
            checks = [
                get_current_state() == condition
                for get_current_state, condition in signal_conditions
            ]
            if check_stopped is True and self.parent.stopped is True:
                return False
            if (all_signals and all(checks)) or (not all_signals and any(checks)):
                return True
            if timer > timeout:
                return False
            time.sleep(interval)
            timer += interval


class PSIDetectorBase(Device):
    """
    Abstract base class for SLS detectors

    Class attributes:
        custom_prepare_cls (object): class for custom prepare logic (BL specific)
        Min_readout (float): minimum readout time for detector

    Args:
        prefix (str): EPICS PV prefix for component (optional)
        name (str): name of the device, as will be reported via read()
        kind (str): member of class 'ophydobj.Kind', defaults to Kind.normal
                    omitted -> readout ignored for read 'ophydobj.read()'
                    normal -> readout for read
                    config -> config parameter for 'ophydobj.read_configuration()'
                    hinted -> which attribute is readout for read
        read_attrs (list): sequence of attribute names to read
        configuration_attrs (list): sequence of attribute names via config_parameters
        parent (object): instance of the parent device
        device_manager (object): bec device manager
        sim_mode (bool): simulation mode, if True, no device manager is required
        **kwargs: keyword arguments

        attributes: lazy_wait_for_connection : bool
    """

    custom_prepare_cls = CustomDetectorMixin

    MIN_READOUT = 1e-3

    # Specify which functions are revealed to the user in BEC client
    USER_ACCESS = ["describe"]

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        device_manager=None,
        sim_mode=False,
        **kwargs,
    ):
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        if device_manager is None and not sim_mode:
            raise DetectorInitError(
                f"No device manager for device: {name}, and not started sim_mode: {sim_mode}. Add"
                " DeviceManager to initialization or init with sim_mode=True"
            )
        # Init variables
        self.sim_mode = sim_mode
        self.stopped = False
        self.name = name
        self.service_cfg = None
        self.scaninfo = None
        self.filewriter = None
        self.timeout = 5
        self.wait_for_connection(all_signals=True)

        # Init custom prepare class with BL specific logic
        self.custom_prepare = self.custom_prepare_cls(parent=self, **kwargs)
        if not sim_mode:
            self._update_service_config()
            self.device_manager = device_manager
        else:
            self.device_manager = bec_utils.DMMock()
            base_path = kwargs["basepath"] if "basepath" in kwargs else "~/Data10/"
            self.service_cfg = {"base_path": os.path.expanduser(base_path)}
        self.producer = self.device_manager.producer
        self._update_scaninfo()
        self._update_filewriter()
        self._init()

    def _update_filewriter(self) -> None:
        """Update filewriter with service config"""
        self.filewriter = FileWriterMixin(self.service_cfg)

    def _update_scaninfo(self) -> None:
        """Update scaninfo from BecScaninfoMixing
        This depends on device manager and operation/sim_mode
        """
        self.scaninfo = BecScaninfoMixin(self.device_manager, self.sim_mode)
        self.scaninfo.load_scan_metadata()

    def _update_service_config(self) -> None:
        """Update service config from BEC service config"""
        self.service_cfg = SERVICE_CONFIG.config["service_config"]["file_writer"]

    def _init(self) -> None:
        """Initialize detector, filewriter and set default parameters"""
        self.custom_prepare.initialize_default_parameter()
        self.custom_prepare.initialize_detector()
        self.custom_prepare.initialize_detector_backend()

    def stage(self) -> list[object]:
        """
         Stage device in preparation for a scan

        Internal Calls:
        - _prep_backend           : prepare detector filewriter for measurement
        - _prep_detector              : prepare detector for measurement

        Returns:
            list(object): list of objects that were staged

        """
        # Method idempotent, should rais ;obj;'RedudantStaging' if staged twice
        if self._staged != Staged.no:
            return super().stage()

        # Reset flag for detector stopped
        self.stopped = False
        # Load metadata of the scan
        self.scaninfo.load_scan_metadata()
        # Prepare detector and file writer
        self.custom_prepare.prepare_detector_backend()
        self.custom_prepare.prepare_detector()
        state = False
        self.custom_prepare.publish_file_location(done=state)
        # At the moment needed bc signal might not be reliable, BEC too fast.
        # Consider removing this overhead in future!
        time.sleep(0.05)
        return super().stage()

    def trigger(self) -> DeviceStatus:
        """Trigger the detector, called from BEC."""
        self.custom_prepare.on_trigger()
        return super().trigger()

    def unstage(self) -> list[object]:
        """
        Unstage device in preparation for a scan

        Returns directly if self.stopped,
        otherwise checks with self._finished
        if data acquisition on device finished (an was successful)

        Internal Calls:
        - custom_prepare.check_scanID          : check if scanID changed or detector stopped
        - custom_prepare.finished              : check if device finished acquisition (succesfully)
        - custom_prepare.publish_file_location : publish file location to bec

        Returns:
            list(object): list of objects that were unstaged
        """
        self.custom_prepare.check_scanID()
        if self.stopped is True:
            return super().unstage()
        self.custom_prepare.finished()
        state = True
        self.custom_prepare.publish_file_location(done=state, successful=state)
        self.stopped = False
        return super().unstage()

    def stop(self, *, success=False) -> None:
        """
        Stop the scan, with camera and file writer

        Internal Calls:
        - custom_prepare.stop_detector     : stop detector
        - custom_prepare.stop_backend : stop detector filewriter
        """
        self.custom_prepare.stop_detector()
        self.custom_prepare.stop_detector_backend()
        super().stop(success=success)
        self.stopped = True



import enum
import functools
import json
import logging
import os
import time

import numpy as np
from prettytable import PrettyTable
from typeguard import typechecked

from ophyd_devices.smaract.smaract_errors import SmaractCommunicationError, SmaractErrorCode
from ophyd_devices.utils.controller import Controller, axis_checked, threadlocked

logger = logging.getLogger("smaract_controller")


class SmaractCommunicationMode(enum.Enum):
    SYNC = 0
    ASYNC = 1


def retry_once(fcn):
    """Decorator to rerun a function in case a SmaractCommunicationError was raised. This may happen if the buffer was not empty."""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        try:
            val = fcn(self, *args, **kwargs)
        except (SmaractCommunicationError, SmaractErrorCode):
            val = fcn(self, *args, **kwargs)
        return val

    return wrapper


class SmaractChannelStatus(enum.Enum):
    STOPPED = 0
    STEPPING = 1
    SCANNING = 2
    HOLDING = 3
    TARGETING = 4
    MOVE_DELAY = 5
    CALIBRATING = 6
    FINDING_REFERENCE_MARK = 7
    LOCKED = 9


class SmaractSensorDefinition:
    def __init__(self, symbol, type_code, positioner_series, comment, reference_type) -> None:
        self.symbol = symbol
        self.type_code = type_code
        self.comment = comment
        self.positioner_series = positioner_series
        self.reference_type = reference_type


class SmaractSensors:
    smaract_sensor_definition_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "smaract_sensors.json"
    )

    def __init__(self) -> None:
        self.avail_sensors = {}

        with open(self.smaract_sensor_definition_file) as json_file:
            sensor_list = json.load(json_file)
        for sensor in sensor_list:
            self.avail_sensors[sensor["type_code"]] = SmaractSensorDefinition(**sensor)


class SmaractController(Controller):
    _axes_per_controller = 6
    _initialized = False
    USER_ACCESS = [
        "socket_put_and_receive",
        "smaract_show_all",
        "move_open_loop_steps",
        "find_reference_mark",
        "describe",
    ]

    def __init__(
        self,
        *,
        name="SmaractController",
        kind=None,
        parent=None,
        socket_cls=None,
        socket_host=None,
        socket_port=None,
        attr_name="",
        labels=None,
    ):
        if not self._initialized:
            super().__init__(
                name=name,
                socket_cls=socket_cls,
                socket_host=socket_host,
                socket_port=socket_port,
                attr_name=attr_name,
                parent=parent,
                labels=labels,
                kind=kind,
            )
            self._sensors = SmaractSensors()

    @threadlocked
    def socket_put(self, val: str):
        self.sock.put(f":{val}\n".encode())

    @threadlocked
    def socket_get(self):
        return self.sock.receive().decode()

    @threadlocked
    def socket_put_and_receive(
        self, val: str, remove_trailing_chars=True, check_for_errors=True, raise_if_not_status=False
    ) -> str:
        self.socket_put(val)
        return_val = ""
        max_wait_time = 1
        elapsed_time = 0
        sleep_time = 0.01
        while True:
            ret = self.socket_get()
            return_val += ret
            if ret.endswith("\n"):
                break
            time.sleep(sleep_time)
            elapsed_time += sleep_time
            if elapsed_time > max_wait_time:
                break
        if remove_trailing_chars:
            return_val = self._remove_trailing_characters(return_val)
            logger.debug(f"Sending {val}; Returned {return_val}")
        if check_for_errors:
            self._check_for_error(return_val, raise_if_not_status=raise_if_not_status)
        return return_val

    @retry_once
    @axis_checked
    def get_status(self, axis_Id_numeric: int) -> SmaractChannelStatus:
        """Returns the current movement status code of a positioner or end effector.This command can be used to check whether a previously issued movement command has been completed.

        Args:
            axis_Id_numeric (int): Axis number

        Returns:
            SmaractChannelStatus: Channel status
        """
        return_val = self.socket_put_and_receive(f"GS{axis_Id_numeric}")
        if self._message_starts_with(return_val, f":S{axis_Id_numeric}"):
            return SmaractChannelStatus(int(return_val.split(",")[1]))

    @retry_once
    @axis_checked
    def is_axis_moving(self, axis_Id_numeric: int) -> bool:
        """Check if axis is moving. Returns true upon open loop move, scanning, closed loop move or reference mark search.

        Args:
            axis_Id_numeric (int): Axis number.

        Returns:
            bool: True if axis is moving.
        """
        axis_status = self.get_status(axis_Id_numeric)
        return axis_status in [
            SmaractChannelStatus.STEPPING,
            SmaractChannelStatus.SCANNING,
            SmaractChannelStatus.TARGETING,
            SmaractChannelStatus.FINDING_REFERENCE_MARK,
        ]

    @retry_once
    def stop_all_axes(self):
        return [
            self.socket_put_and_receive(f"S{ax.axis_Id_numeric}", raise_if_not_status=True)
            for ax in self._axis
            if ax is not None
        ]

    @retry_once
    @axis_checked
    def axis_is_referenced(self, axis_Id_numeric: int) -> bool:
        return_val = self.socket_put_and_receive(f"GPPK{axis_Id_numeric}")
        if self._message_starts_with(return_val, f":PPK{axis_Id_numeric}"):
            return bool(int(return_val.split(",")[1]))

    def all_axes_referenced(self) -> bool:
        return all(self.axis_is_referenced(ax) for ax in self._axis if ax is not None)

    @retry_once
    @axis_checked
    def get_position(self, axis_Id_numeric: int) -> float:
        """Returns the current position of a positioner.

        Args:
            axis_Id_numeric (int): Axis number.

        Returns:
            float: Position in mm
        """
        return_val = self.socket_put_and_receive(f"GP{axis_Id_numeric}")
        if self._message_starts_with(return_val, f":P{axis_Id_numeric}"):
            return float(return_val.split(",")[1]) / 1e6

    @retry_once
    @axis_checked
    @typechecked
    def move_axis_to_absolute_position(
        self, axis_Id_numeric: int, target_val: float, hold_time: int = 1000
    ) -> None:
        """Instructs a positioner to move to a specific position.

        Args:
            axis_Id_numeric (int): Axis number.
            target_val (float): Target position in mm.
            hold_time (int, optional): Specifies how long (in milliseconds) the position is actively held after reaching the target. The valid range is 0..60,000. A 0 deactivates this feature, a value of 60,000 is infinite (until manually stopped, see S command). Defaults to 1000.

        """
        self.socket_put_and_receive(
            f"MPA{axis_Id_numeric},{int(np.round(target_val*1e6))},{hold_time}",
            raise_if_not_status=True,
        )

    @retry_once
    @axis_checked
    @typechecked
    def move_axis_to_relative_position(
        self, axis_Id_numeric: int, target_val: float, hold_time: int = 1000
    ) -> None:
        """Instructs a positioner to move to a position relative to its current position.

        Args:
            axis_Id_numeric (int): Axis number.
            target_val (float): Relative position to move to in mm.
            hold_time (int, optional): Specifies how long (in milliseconds) the position is actively held after reaching the target. The valid range is 0..60,000. A 0 deactivates this feature, a value of 60,000 is infinite (until manually stopped, see S command). Defaults to 1000.

        """
        self.socket_put_and_receive(
            f"MPR{axis_Id_numeric},{int(np.round(target_val*1e6))},{hold_time}",
            raise_if_not_status=True,
        )

    @retry_once
    @axis_checked
    @typechecked
    def move_open_loop_steps(
        self, axis_Id_numeric: int, steps: int, amplitude: int = 4000, frequency: int = 2000
    ) -> None:
        """Move open loop steps. It performs a burst of steps with the given parameters.

        Args:
            axis_Id_numeric (int): Axis number.
            steps (int): Number and direction of steps to perform. The valid range is -30,000..30,000. A value of 0 stops the positioner, but see S command. A value of 30,000 or -30,000 performs an unbounded move. This should be used with caution since the positioner will only stop on an S command.
            amplitude (int): Amplitude that the steps are performed with. Lower amplitude values result in a smaller step width. The parameter must be given as a 12bit value (range 0..4,095). 0 corresponds to 0V, 4,095 to 100V. Default: 4000
            frequency (int): Frequency in Hz that the steps are performed with. The valid range is 1..18,500. Default: 2000.
        """
        self.socket_put_and_receive(
            f"MST{axis_Id_numeric},{steps},{amplitude},{frequency}", raise_if_not_status=True
        )

    @retry_once
    def get_communication_mode(self) -> SmaractCommunicationMode:
        return_val = self.socket_put_and_receive("GCM")
        if self._message_starts_with(return_val, f":CM"):
            return SmaractCommunicationMode(int(return_val.strip(":CM")))

    @retry_once
    @axis_checked
    def get_channel_type(self, axis_Id_numeric) -> str:
        return_val = self.socket_put_and_receive(f"GCT{axis_Id_numeric}")
        if self._message_starts_with(return_val, f":CT{axis_Id_numeric}"):
            return return_val.split(",")[1]

    @retry_once
    def get_interface_version(self) -> str:
        """This command may be used to retrieve the interface version of the system. It is useful to check if changes
        have been made to the software interface. An application may check the version in order to ensure that the
        system behaves as the application expects it to do.

        Returns:
            str: interface version
        """
        return_val = self.socket_put_and_receive("GIV")
        if self._message_starts_with(return_val, f":IV"):
            return return_val.strip(":IV")

    @retry_once
    def get_number_of_channels(self) -> int:
        """This command may be used to determine how many control channels are available on a system. This
        includes positioner channels and end effector channels. Each channel is of a specific type. Use the GCT
        command to determine the types of the channels.
        Note that the number of channels does not represent the number positioners and/or end effectors that are
        currently connected to the system.
        The channel indexes throughout the interface are zero based. If your system has N channels then the valid
        range for a channel index is 0.. N-1.

        Returns:
            int: number of channels
        """
        return_val = self.socket_put_and_receive("GNC")
        if self._message_starts_with(return_val, f":N"):
            return int(return_val.strip(":N"))

    @retry_once
    def get_system_id(self) -> str:
        """This command may be used to physically identify a system connected to the PC. Each system has a unique
        ID which makes it possible to distinguish one from another.
        The ID returned is a generic decimal number that uniquely identifies the system.

        """
        return_val = self.socket_put_and_receive("GSI")
        if self._message_starts_with(return_val, f":ID"):
            return return_val.strip(":ID")

    @retry_once
    def reset(self) -> None:
        """When this command is sent the system will perform a reset. It has the same effect as a power down/power
        up cycle. The system replies with an acknowledge string before resetting itself.
        """
        self.socket_put_and_receive("R", raise_if_not_status=True)

    @retry_once
    def set_hcm_mode(self, mode: int):
        """If a Hand Control Module (HCM) is connected to the system, this command may be used to enable or
        disable it in order to avoid interference while the software is in control of the system. There are three possible
        modes to set:
        0: In this mode the Hand Control Module is disabled. It may not be used to control positioners.
        1: This is the default setting where the Hand Control Module may be used to control the positioners.
        2: In this mode the Hand Control Module cannot be used to control the positioners. However, if there
        are positioners with sensors attached, their position data will still be displayed.

        Args:
            mode (int): HCM mode

        """
        if mode not in range(3):
            raise ValueError(f"HCM mode must be 0, 1 or 2. Received: {mode}.")
        self.socket_put_and_receive(f"SHE{mode}", raise_if_not_status=True)

    @retry_once
    @axis_checked
    def get_position_limits(self, axis_Id_numeric: int) -> list:
        """May be used to read out the travel range limit that is currently
        configured for a linear channel.

                Args:
                    axis_Id_numeric (int): Axis

                Returns:
                    list: [low_limit, high_limit] in mm
        """
        return_val = self.socket_put_and_receive(f"GPL{axis_Id_numeric}")
        if self._message_starts_with(return_val, f":GPL{axis_Id_numeric}"):
            return [
                float(limit) / 1e6
                for limit in return_val.strip(f":GPL{axis_Id_numeric},").split(",")
            ]

    @retry_once
    @axis_checked
    def set_position_limits(
        self, axis_Id_numeric: int, low_limit: float, high_limit: float
    ) -> None:
        """For positioners with integrated sensors this command may be used to limit the travel range of a linear
        positioner by software. By default there is no limit set. If defined the
        positioner will not move beyond the limit. This affects open-loop as well as closed-loop movements.

                Args:
                    axis_Id_numeric (int): Axis
                    low_limit (float): low limit in mm
                    high_limit (float): high limit in mm

        """
        self.socket_put_and_receive(
            f"SPL{axis_Id_numeric},{np.round(low_limit*1e6)},{np.round(high_limit*1e6)}",
            raise_if_not_status=True,
        )

    @retry_once
    @axis_checked
    def get_sensor_type(self, axis_Id_numeric: int) -> SmaractSensorDefinition:
        return_val = self.socket_put_and_receive(f"GST{axis_Id_numeric}")
        if self._message_starts_with(return_val, f":ST{axis_Id_numeric}"):
            return self._sensors.avail_sensors.get(int(return_val.strip(f":ST{axis_Id_numeric},")))

    @retry_once
    @axis_checked
    def find_reference_mark(
        self, axis_Id_numeric: int, direction: int, holdTime: int, autoZero: int
    ) -> None:
        return_val = self.socket_put_and_receive(
            f"FRM{axis_Id_numeric},{direction},{holdTime},{autoZero}"
        )

    @retry_once
    @axis_checked
    def set_closed_loop_move_speed(self, axis_Id_numeric: int, move_speed: float) -> None:
        """This command configures the speed control feature of a channel for closed-loop commands move_axis_to_absolute_position. By default the speed control is inactive. In this state the behavior of closed-loop commands is influenced by the maximum driving frequency. If a movement speed is configured, all following closed-loop commands will be executed with the new speed.

        Args:
            axis_Id_numeric (int): Axis number.
            move_speed (float): Movement speed given in mm/s for linear positioners. The valid range is 0 .. 100. A value of 0 (default) deactivates the speed control feature.
        """
        move_speed_in_nm_per_s = int(round(move_speed * 1e6))

        if move_speed_in_nm_per_s > 100e6 or move_speed_in_nm_per_s < 0:
            raise ValueError("Move speed must be within 0 to 100 mm/s.")

        self.socket_put_and_receive(
            f"SCLS{axis_Id_numeric},{move_speed_in_nm_per_s}", raise_if_not_status=True
        )

    @retry_once
    @axis_checked
    def get_closed_loop_move_speed(self, axis_Id_numeric: int) -> float:
        """Returns the currently configured movement speed that is used for closed-loop commands for a channel.

        Args:
            axis_Id_numeric (int): Axis number.

        Returns:
            float: move speed in mm/s. A return value of 0 means that the speed control feature is disabled.
        """

        return_val = self.socket_put_and_receive(f"GCLS{axis_Id_numeric}")
        if self._message_starts_with(return_val, f":CLS{axis_Id_numeric}"):
            return float(return_val.strip(f":CLS{axis_Id_numeric},")) * 1e6

    def describe(self) -> None:
        t = PrettyTable()
        t.title = f"{self.__class__.__name__} on {self.sock.host}:{self.sock.port}"
        t.field_names = ["Axis", "Name", "Connected", "Referenced", "Closed Loop Speed", "Position"]
        for ax in range(self._axes_per_controller):
            axis = self._axis[ax]
            if axis is not None:
                t.add_row(
                    [
                        f"{axis.axis_Id_numeric}/{axis.axis_Id}",
                        axis.name,
                        axis.connected,
                        self.axis_is_referenced(axis.axis_Id_numeric),
                        self.get_closed_loop_move_speed(axis.axis_Id_numeric),
                        axis.readback.read().get(axis.name).get("value"),
                    ]
                )
            else:
                t.add_row([None for t in t.field_names])
        print(t)

    @axis_checked
    def _error_str(self, axis_Id_numeric: int, error_number: int):
        return f":E{axis_Id_numeric},{error_number}"

    def _get_error_code_from_msg(self, msg: str) -> int:
        if msg.startswith(":E"):
            return int(msg.split(",")[-1])
        else:
            return -1

    def _get_axis_from_error_code(self, msg: str) -> int:
        if msg.startswith(":E"):
            try:
                return int(msg.strip(":E").split(",")[0])
            except ValueError:
                return None
        else:
            return None

    def _check_for_error(self, msg: str, axis_Id_numeric: int = None, raise_if_not_status=False):
        if msg.startswith(":E"):
            if axis_Id_numeric is None:
                axis_Id_numeric = self._get_axis_from_error_code(msg)

            if axis_Id_numeric is None:
                raise SmaractCommunicationError(
                    "Could not retrieve axis number from error message."
                )

            if msg != self._error_str(axis_Id_numeric, 0):
                error_code = self._get_error_code_from_msg(msg)
                if error_code != 0:
                    raise SmaractErrorCode(error_code)
        else:
            if raise_if_not_status:
                raise SmaractCommunicationError(
                    "Expected error / status message but failed to parse it."
                )

    def _remove_trailing_characters(self, var: str) -> str:
        if len(var) > 1:
            return var.split("\n")[0]
        return var

    def _message_starts_with(self, msg: str, leading_chars: str) -> bool:
        if msg.startswith(leading_chars):
            return True
        raise SmaractCommunicationError(
            f"Expected to receive a return message starting with {leading_chars} but instead"
            f" received '{msg}'"
        )



class BECDeviceBase:
    """Base class for BEC devices with zero functionality."""

    def __init__(self, name: str, *args, **kwargs):
        self.name = name
        self._connected = True
        self._destroyed = False

    @property
    def hints(self):
        """hints property"""
        return {}

    @property
    def connected(self) -> bool:
        """connected property.
        Check if signals are connected

        Returns:
            bool: True if connected, False otherwise
        """
        return self._connected

    @connected.setter
    def connected(self, value: bool):
        """connected setter"""
        self._connected = value

    def describe(self) -> dict:
        """describe method

        Includes all signals of type Kind.hinted and Kind.normal.
        Override by child class with describe method

        Returns:
            dict: Dictionary with dictionaries with signal descriptions ('source', 'dtype', 'shape')
        """
        return {}

    def describe_configuration(self) -> dict:
        """describe method

        Includes all signals of type Kind.config.
        Override by child class with describe_configuration method

        Returns:
            dict: Dictionary with dictionaries with signal descriptions ('source', 'dtype', 'shape')
        """
        return {}

    def read_configuration(self) -> dict:
        """read_configuration method

        Override by child class with read_configuration method

        Returns:
            dict: Dictionary with nested dictionary of signals with kind.config:
            {'signal_name' : {'value' : .., "timestamp" : ..}, ...}
        """
        return {}

    def read(self) -> dict:
        """read method

        Override by child class with read method

        Returns:
            dict: Dictionary with nested dictionary of signals with kind.normal or kind.hinted:
            {'signal_name' : {'value' : .., "timestamp" : ..}, ...}
        """
        return {}

    def destroy(self):
        """Destroy method"""
        self._destroyed = True



import threading
import time
from typing import List

import numpy as np
from bec_lib import MessageEndpoints, bec_logger, messages
from ophyd import Component as Cpt
from ophyd import Device, PositionerBase, Signal
from ophyd.status import wait as status_wait
from ophyd.utils import LimitError
from prettytable import PrettyTable

from ophyd_devices.rt_lamni.rt_ophyd import (
    BECConfigError,
    RtCommunicationError,
    RtController,
    RtError,
    RtReadbackSignal,
    RtSetpointSignal,
    RtSignalRO,
    retry_once,
)
from ophyd_devices.utils.controller import threadlocked
from ophyd_devices.utils.socket import SocketIO, raise_if_disconnected

logger = bec_logger.logger


class RtFlomniController(RtController):
    USER_ACCESS = [
        "socket_put_and_receive",
        "set_rotation_angle",
        "feedback_disable",
        "feedback_enable_without_reset",
        "feedback_enable_with_reset",
        "add_pos_to_scan",
        "clear_trajectory_generator",
        "show_cyclic_error_compensation",
        "laser_tracker_on",
        "laser_tracker_off",
        "laser_tracker_show_all",
    ]

    def __init__(
        self,
        *,
        name=None,
        socket_cls=None,
        socket_host=None,
        socket_port=None,
        attr_name="",
        parent=None,
        labels=None,
        kind=None,
    ):
        super().__init__(
            name=name,
            socket_cls=socket_cls,
            socket_host=socket_host,
            socket_port=socket_port,
            attr_name=attr_name,
            parent=parent,
            labels=labels,
            kind=kind,
        )
        self.tracker_info = {}
        self._min_scan_buffer_reached = False
        self.rt_pid_voltage = None

    def add_pos_to_scan(self, positions) -> None:
        def send_positions(parent, positions):
            parent._min_scan_buffer_reached = False
            start_time = time.time()
            for pos_index, pos in enumerate(positions):
                parent.socket_put_and_receive(f"s{pos[0]},{pos[1]},{pos[2]}")
                if pos_index > 100:
                    parent._min_scan_buffer_reached = True
            parent._min_scan_buffer_reached = True
            logger.info(
                f"Sending {len(positions)} positions took {time.time()-start_time} seconds."
            )

        threading.Thread(target=send_positions, args=(self, positions), daemon=True).start()

    def move_to_zero(self):
        self.socket_put("pa0,0")
        self.get_axis_by_name("rtx").user_setpoint.setpoint = 0
        self.socket_put("pa1,0")
        self.get_axis_by_name("rty").user_setpoint.setpoint = 0
        self.socket_put("pa2,0")
        self.get_axis_by_name("rtz").user_setpoint.setpoint = 0
        time.sleep(0.05)

    def feedback_is_running(self) -> bool:
        status = int(float(self.socket_put_and_receive("l2").strip()))
        if status == 1:
            return False
        return True

    def feedback_enable_with_reset(self):
        self.socket_put("l0")  # disable feedback

        self.move_to_zero()

        if not self.slew_rate_limiters_on_target() or np.abs(self.pid_y()) > 0.1:
            logger.info("Please wait, slew rate limiters not on target.")
        while not self.slew_rate_limiters_on_target() or np.abs(self.pid_y()) > 0.1:
            time.sleep(0.05)

        self.get_device_manager().devices.rty.update_user_parameter({"tomo_additional_offsety": 0})
        self.clear_trajectory_generator()

        self.laser_tracker_on()

        # move to 0. FUPR will set the rotation angle during readout
        self.get_device_manager().devices.fsamroy.obj.move(0, wait=True)

        fsamx = self.get_device_manager().devices.fsamx

        fsamx.obj.pid_x_correction = 0
        fsamx.obj.controller.socket_put_confirmed("axspeed[4]=0.1*stppermm[4]")
        fsamx_in = fsamx.user_parameter.get("in")
        if not np.isclose(fsamx.obj.readback.get(), fsamx_in, atol=0.3):
            raise RtError(
                "Something is wrong. fsamx is very far from the samx_in position. Don't dare correct automatically."
            )

        if not np.isclose(fsamx.obj.readback.get(), fsamx_in, atol=0.01):
            fsamx.enabled_set = True
            fsamx.obj.move(fsamx_in, wait=True)
            fsamx.enabled_set = False
            time.sleep(1)

        self.socket_put("l1")
        time.sleep(0.4)

        if not self.feedback_is_running():
            raise RtError("Feedback is not running; likely an error in the interferometer.")

        time.sleep(1.5)
        self.show_cyclic_error_compensation()

        self.rt_pid_voltage = self.get_pid_x()
        rtx = self.get_device_manager().devices.rtx
        rtx.update_user_parameter({"rt_pid_voltage": self.rt_pid_voltage})

        self.set_device_enabled("fsamx", False)
        self.set_device_enabled("fsamy", False)
        self.set_device_enabled("foptx", False)
        self.set_device_enabled("fopty", False)

    def move_samx_to_scan_region(self, fovx: float, cenx: float):
        if self.rt_pid_voltage is None:
            rtx = self.get_device_manager().devices.rtx
            self.rt_pid_voltage = rtx.user_parameter.get("rt_pid_voltage")
            if self.rt_pid_voltage is None:
                raise RtError(
                    "rt_pid_voltage not set in rtx user parameters. Please run feedback_enable_with_reset first."
                )
            logger.info(f"Using PID voltage from rtx user parameter: {self.rt_pid_voltage}")
        expected_voltage = self.rt_pid_voltage + fovx / 2 * 7 / 100
        logger.info(f"Expected PID voltage: {expected_voltage}")
        logger.info(f"Current PID voltage: {self.get_pid_x()}")

        wait_on_exit = False
        while True:
            if np.abs(self.get_pid_x() - expected_voltage) < 1:
                break
            wait_on_exit = True
            self.socket_put("v0")
            fsamx = self.get_device_manager().devices.fsamx
            fsamx.enabled_set = True
            fsamx.obj.controller.socket_put_confirmed("axspeed[4]=0.1*stppermm[4]")
            fsamx.obj.pid_x_correction -= (self.get_pid_x() - expected_voltage) * 0.007
            logger.info(f"Correcting fsamx by {fsamx.obj.pid_x_correction}")
            fsamx_in = fsamx.user_parameter.get("in")
            fsamx.obj.move(fsamx_in + cenx / 1000 + fsamx.obj.pid_x_correction, wait=True)
            fsamx.enabled_set = False
            time.sleep(0.1)
            self.laser_tracker_on()
            time.sleep(0.01)

        if wait_on_exit:
            time.sleep(1)

        self.socket_put("v1")

    @threadlocked
    def clear_trajectory_generator(self):
        self.socket_put("sc")
        logger.info("flomni scan stopped and deleted, moving to start position")

    def feedback_enable_without_reset(self):
        self.laser_tracker_on()
        self.socket_put("l3")
        time.sleep(0.01)

        if not self.feedback_is_running():
            raise RtError("Feedback is not running; likely an error in the interferometer.")

        self.set_device_enabled("fsamx", False)
        self.set_device_enabled("fsamy", False)
        self.set_device_enabled("foptx", False)
        self.set_device_enabled("fopty", False)

    def feedback_disable(self):
        self.clear_trajectory_generator()
        self.move_to_zero()
        self.socket_put("l0")

        self.set_device_enabled("fsamx", True)
        self.set_device_enabled("fsamy", True)
        self.set_device_enabled("foptx", True)
        self.set_device_enabled("fopty", True)

        fsamx = self.get_device_manager().devices.fsamx
        fsamx.obj.controller.socket_put_confirmed("axspeed[4]=025*stppermm[4]")

    def get_pid_x(self) -> float:
        voltage = float(self.socket_put_and_receive("g").strip())
        return voltage

    def show_cyclic_error_compensation(self):
        cec0 = int(float(self.socket_put_and_receive("w0").strip()))
        cec1 = int(float(self.socket_put_and_receive("w0").strip()))

        if cec0 == 32:
            logger.info("Cyclic Error Compensation: y-axis is initialized")
        else:
            logger.info("Cyclic Error Compensation: y-axis is NOT initialized")
        if cec1 == 32:
            logger.info("Cyclic Error Compensation: x-axis is initialized")
        else:
            logger.info("Cyclic Error Compensation: y-axis is NOT initialized")

    def set_rotation_angle(self, val: float) -> None:
        self.socket_put(f"a{val/180*np.pi}")

    def laser_tracker_on(self):
        self.laser_update_tracker_info()

        if not self.tracker_info["enabled_z"] or not self.tracker_info["enabled_y"]:
            logger.info("Enabling the laser tracker. Please wait...")

            tracker_intensity = self.tracker_info["tracker_intensity"]
            if (
                tracker_intensity < self.tracker_info["threshold_intensity_y"]
                or tracker_intensity < self.tracker_info["threshold_intensity_z"]
            ):
                logger.info(self.tracker_info)
                raise RtError("The tracker cannot be enabled because the beam intensity it low.")

            self.move_to_zero()
            self.socket_put("T1")
            time.sleep(0.5)

            self.get_device_manager().devices.ftrackz.obj.controller.socket_put_confirmed(
                "trackyct=0"
            )
            self.get_device_manager().devices.ftrackz.obj.controller.socket_put_confirmed(
                "trackzct=0"
            )

        self.laser_tracker_wait_on_target()
        logger.info("Laser tracker running!")

    def laser_tracker_off(self):
        self.socket_put("T0")

    def laser_tracker_show_all(self):
        self.laser_update_tracker_info()
        t = PrettyTable()
        t.title = f"Laser Tracker Info"
        t.field_names = ["Name", "Value"]
        for key, val in self.tracker_info.items():
            t.add_row([key, val])
        print(t)

    def laser_update_tracker_info(self):
        ret = self.socket_put_and_receive("Ts")

        # remove trailing \n
        ret = ret.split("\n")[0]

        tracker_values = [float(val) for val in ret.split(",")]
        self.tracker_info = {
            "tracker_intensity": tracker_values[2],
            "threshold_intensity_y": tracker_values[8],
            "enabled_y": bool(tracker_values[10]),
            "beampos_y": tracker_values[5],
            "target_y": tracker_values[6],
            "piezo_voltage_y": tracker_values[9],
            "threshold_intensity_z": tracker_values[3],
            "enabled_z": bool(tracker_values[10]),
            "beampos_z": tracker_values[0],
            "target_z": tracker_values[1],
            "piezo_voltage_z": tracker_values[4],
        }

    def laser_tracker_galil_enable(self):
        self.get_device_manager().devices.ftrackz.obj.controller.socket_put_confirmed("tracken=1")
        self.get_device_manager().devices.ftrackz.obj.controller.socket_put_confirmed("trackyct=0")
        self.get_device_manager().devices.ftrackz.obj.controller.socket_put_confirmed("trackzct=0")
        self.get_device_manager().devices.ftrackz.obj.controller.socket_put_confirmed("XQ#Tracker")

    def laser_tracker_on_target(self) -> bool:
        self.laser_update_tracker_info()
        if np.isclose(
            self.tracker_info["beampos_y"], self.tracker_info["target_y"], atol=0.02
        ) and np.isclose(self.tracker_info["beampos_z"], self.tracker_info["target_z"], atol=0.02):
            return True
        return False

    def laser_tracker_wait_on_target(self):
        max_repeat = 25
        count = 0
        while not self.laser_tracker_on_target():
            self.laser_tracker_galil_enable()
            logger.info("Waiting for laser tracker to reach target.")
            time.sleep(0.5)
            count += 1
            if count > max_repeat:
                raise RtError("Failed to reach laser target position.")

    def slew_rate_limiters_on_target(self) -> bool:
        ret = int(float(self.socket_put_and_receive("y").strip()))
        if ret == 3:
            return True
        return False

    def pid_y(self) -> float:
        ret = float(self.socket_put_and_receive("G").strip())
        return ret

    def read_ssi_interferometer(self, axis_number):
        val = float(self.socket_put_and_receive(f"j{axis_number}").strip())
        return val

    def show_signal_strength_interferometer(self):
        t = PrettyTable()
        t.title = f"Interferometer signal strength"
        t.field_names = ["Axis", "Value"]
        for i in range(3):
            t.add_row([i, self.read_ssi_interferometer(i)])
        print(t)

    def _get_signals_from_table(self, return_table) -> dict:
        self.average_stdeviations_x_st_fzp += float(return_table[4])
        self.average_stdeviations_y_st_fzp += float(return_table[7])
        signals = {
            "target_x": {"value": float(return_table[2])},
            "average_x_st_fzp": {"value": float(return_table[3])},
            "stdev_x_st_fzp": {"value": float(return_table[4])},
            "target_y": {"value": float(return_table[5])},
            "average_y_st_fzp": {"value": float(return_table[6])},
            "stdev_y_st_fzp": {"value": float(return_table[7])},
            "average_rotz": {"value": float(return_table[8])},
            "stdev_rotz": {"value": float(return_table[9])},
            "average_stdeviations_x_st_fzp": {
                "value": self.average_stdeviations_x_st_fzp / (int(return_table[0]) + 1)
            },
            "average_stdeviations_y_st_fzp": {
                "value": self.average_stdeviations_y_st_fzp / (int(return_table[0]) + 1)
            },
        }
        return signals

    @threadlocked
    def start_scan(self):
        if not self.feedback_is_running():
            logger.error(
                "Cannot start scan because feedback loop is not running or there is an"
                " interferometer error."
            )
            raise RtError(
                "Cannot start scan because feedback loop is not running or there is an"
                " interferometer error."
            )
            # here exception
        (mode, number_of_positions_planned, current_position_in_scan) = self.get_scan_status()

        if number_of_positions_planned == 0:
            logger.error("Cannot start scan because no target positions are planned.")
            raise RtError("Cannot start scan because no target positions are planned.")
            # hier exception
        # start a point-by-point scan (for cont scan in flomni it would be "sa")
        self.socket_put_and_receive("sd")

    @retry_once
    @threadlocked
    def get_scan_status(self):
        return_table = (self.socket_put_and_receive("sr")).split(",")
        if len(return_table) != 3:
            raise RtCommunicationError(
                f"Expected to receive 3 return values. Instead received {return_table}"
            )
        mode = int(float(return_table[0]))
        # mode 0: direct positioning
        # mode 1: running internal timer (not tested/used anymore)
        # mode 2: rt point scan running
        # mode 3: rt point scan starting
        # mode 5/6: rt continuous scanning (not available in LamNI)
        number_of_positions_planned = int(float(return_table[1]))
        current_position_in_scan = int(float(return_table[2]))
        return (mode, number_of_positions_planned, current_position_in_scan)

    def get_device_manager(self):
        for axis in self._axis:
            if hasattr(axis, "device_manager") and axis.device_manager:
                return axis.device_manager
        raise BECConfigError("Could not access the device_manager")

    def read_positions_from_sampler(self):
        # this was for reading after the scan completed
        number_of_samples_to_read = 1  # self.get_scan_status()[1]  #number of valid samples, will be updated upon first data read

        read_counter = 0

        self.average_stdeviations_x_st_fzp = 0
        self.average_stdeviations_y_st_fzp = 0
        self.average_lamni_angle = 0

        mode, number_of_positions_planned, current_position_in_scan = self.get_scan_status()

        # if not (mode==2 or mode==3):
        #    error
        self.get_device_manager().producer.set_and_publish(
            MessageEndpoints.device_status("rt_scan"),
            messages.DeviceStatusMessage(
                device="rt_scan", status=1, metadata=self.readout_metadata
            ).dumps(),
        )
        # while scan is running
        while mode > 0:
            # logger.info(f"Current scan position {current_position_in_scan} out of {number_of_positions_planned}")
            mode, number_of_positions_planned, current_position_in_scan = self.get_scan_status()
            time.sleep(0.01)
            if current_position_in_scan > 5:
                while current_position_in_scan > read_counter + 1:
                    return_table = (self.socket_put_and_receive(f"r{read_counter}")).split(",")
                    # logger.info(f"{return_table}")
                    logger.info(f"Read {read_counter} out of {number_of_positions_planned}")

                    read_counter = read_counter + 1

                    signals = self._get_signals_from_table(return_table)

                    self.publish_device_data(signals=signals, pointID=int(return_table[0]))

        time.sleep(0.05)

        # read the last samples even though scan is finished already
        while number_of_positions_planned > read_counter:
            return_table = (self.socket_put_and_receive(f"r{read_counter}")).split(",")
            logger.info(f"Read {read_counter} out of {number_of_positions_planned}")
            # logger.info(f"{return_table}")
            read_counter = read_counter + 1

            signals = self._get_signals_from_table(return_table)
            self.publish_device_data(signals=signals, pointID=int(return_table[0]))

        self.get_device_manager().producer.set_and_publish(
            MessageEndpoints.device_status("rt_scan"),
            messages.DeviceStatusMessage(
                device="rt_scan", status=0, metadata=self.readout_metadata
            ).dumps(),
        )

        logger.info(
            "Flomni statistics: Average of all standard deviations: x"
            f" {self.average_stdeviations_x_st_fzp/number_of_samples_to_read}, y"
            f" {self.average_stdeviations_y_st_fzp/number_of_samples_to_read}."
        )

    def publish_device_data(self, signals, pointID):
        self.get_device_manager().producer.send(
            MessageEndpoints.device_read("rt_flomni"),
            messages.DeviceMessage(
                signals=signals, metadata={"pointID": pointID, **self.readout_metadata}
            ).dumps(),
        )

    def start_readout(self):
        readout = threading.Thread(target=self.read_positions_from_sampler)
        readout.start()

    def kickoff(self, metadata):
        self.readout_metadata = metadata
        while not self._min_scan_buffer_reached:
            time.sleep(0.001)
        self.start_scan()
        time.sleep(0.1)
        self.start_readout()


class RtFlomniReadbackSignal(RtReadbackSignal):
    @retry_once
    @threadlocked
    def _socket_get(self) -> float:
        """Get command for the readback signal

        Returns:
        float: Readback value after adjusting for sign and motor resolution.
        """
        time.sleep(0.1)
        return_table = (self.controller.socket_put_and_receive(f"pr")).split(",")

        current_pos = float(return_table[self.parent.axis_Id_numeric])

        current_pos *= self.parent.sign
        self.parent.user_setpoint.setpoint = current_pos
        return current_pos


class RtFlomniSetpointSignal(RtSetpointSignal):
    setpoint = 0

    @retry_once
    @threadlocked
    def _socket_set(self, val: float) -> None:
        """Set a new target value / setpoint value. Before submission, the target value is adjusted for the axis' sign.
        Furthermore, it is ensured that all axes are referenced before a new setpoint is submitted.

        Args:
            val (float): Target value / setpoint value

        Raises:
            RtError: Raised if interferometer feedback is disabled.

        """
        if not self.parent.controller.feedback_is_running():
            raise RtError(
                "The interferometer feedback is not running. Either it is turned off or and"
                " interferometer error occured."
            )
        self.set_with_feedback_disabled(val)

    def set_with_feedback_disabled(self, val):
        target_val = val * self.parent.sign
        self.setpoint = target_val
        self.controller.socket_put(f"pa{self.parent.axis_Id_numeric},{target_val:.4f}")


class RtFlomniFeedbackRunning(RtSignalRO):
    @threadlocked
    def _socket_get(self):
        return int(self.parent.controller.feedback_is_running())


class RtFlomniMotor(Device, PositionerBase):
    USER_ACCESS = ["controller"]
    readback = Cpt(RtFlomniReadbackSignal, signal_name="readback", kind="hinted")
    user_setpoint = Cpt(RtFlomniSetpointSignal, signal_name="setpoint")

    high_limit_travel = Cpt(Signal, value=0, kind="omitted")
    low_limit_travel = Cpt(Signal, value=0, kind="omitted")

    SUB_READBACK = "readback"
    SUB_CONNECTION_CHANGE = "connection_change"
    _default_sub = SUB_READBACK

    def __init__(
        self,
        axis_Id,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        host="mpc2844.psi.ch",
        port=2222,
        sign=1,
        socket_cls=SocketIO,
        device_manager=None,
        limits=None,
        **kwargs,
    ):
        self.axis_Id = axis_Id
        self.sign = sign
        self.controller = RtFlomniController(
            socket_cls=socket_cls, socket_host=host, socket_port=port
        )
        self.controller.set_axis(axis=self, axis_nr=self.axis_Id_numeric)
        self.device_manager = device_manager
        self.tolerance = kwargs.pop("tolerance", 0.5)

        super().__init__(
            prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            **kwargs,
        )
        self.readback.name = self.name
        self.controller.subscribe(
            self._update_connection_state, event_type=self.SUB_CONNECTION_CHANGE
        )
        self._update_connection_state()

        # self.readback.subscribe(self._forward_readback, event_type=self.readback.SUB_VALUE)
        if limits is not None:
            assert len(limits) == 2
            self.low_limit_travel.put(limits[0])
            self.high_limit_travel.put(limits[1])

    @property
    def limits(self):
        return (self.low_limit_travel.get(), self.high_limit_travel.get())

    @property
    def low_limit(self):
        return self.limits[0]

    @property
    def high_limit(self):
        return self.limits[1]

    def check_value(self, pos):
        """Check that the position is within the soft limits"""
        low_limit, high_limit = self.limits

        if low_limit < high_limit and not (low_limit <= pos <= high_limit):
            raise LimitError(f"position={pos} not within limits {self.limits}")

    def _update_connection_state(self, **kwargs):
        for walk in self.walk_signals():
            walk.item._metadata["connected"] = self.controller.connected

    def _forward_readback(self, **kwargs):
        kwargs.pop("sub_type")
        self._run_subs(sub_type="readback", **kwargs)

    @raise_if_disconnected
    def move(self, position, wait=True, **kwargs):
        """Move to a specified position, optionally waiting for motion to
        complete.

        Parameters
        ----------
        position
            Position to move to
        moved_cb : callable
            Call this callback when movement has finished. This callback must
            accept one keyword argument: 'obj' which will be set to this
            positioner instance.
        timeout : float, optional
            Maximum time to wait for the motion. If None, the default timeout
            for this positioner is used.

        Returns
        -------
        status : MoveStatus

        Raises
        ------
        TimeoutError
            When motion takes longer than `timeout`
        ValueError
            On invalid positions
        RuntimeError
            If motion fails other than timing out
        """
        self._started_moving = False
        timeout = kwargs.pop("timeout", 100)
        status = super().move(position, timeout=timeout, **kwargs)
        self.user_setpoint.put(position, wait=False)

        def move_and_finish():
            while not self.controller.slew_rate_limiters_on_target():
                print("motor is moving")
                val = self.readback.read()
                self._run_subs(sub_type=self.SUB_READBACK, value=val, timestamp=time.time())
                time.sleep(0.01)
            print("Move finished")
            self._done_moving()

        threading.Thread(target=move_and_finish, daemon=True).start()
        try:
            if wait:
                status_wait(status)
        except KeyboardInterrupt:
            self.stop()
            raise

        return status

    @property
    def axis_Id(self):
        return self._axis_Id_alpha

    @axis_Id.setter
    def axis_Id(self, val):
        if isinstance(val, str):
            if len(val) != 1:
                raise ValueError(f"Only single-character axis_Ids are supported.")
            self._axis_Id_alpha = val
            self._axis_Id_numeric = ord(val.lower()) - 97
        else:
            raise TypeError(f"Expected value of type str but received {type(val)}")

    @property
    def axis_Id_numeric(self):
        return self._axis_Id_numeric

    @axis_Id_numeric.setter
    def axis_Id_numeric(self, val):
        if isinstance(val, int):
            if val > 26:
                raise ValueError(f"Numeric value exceeds supported range.")
            self._axis_Id_alpha = val
            self._axis_Id_numeric = (chr(val + 97)).capitalize()
        else:
            raise TypeError(f"Expected value of type int but received {type(val)}")

    def kickoff(self, metadata, **kwargs) -> None:
        self.controller.kickoff(metadata)

    @property
    def egu(self):
        """The engineering units (EGU) for positions"""
        return "um"

    # how is this used later?

    def stage(self) -> List[object]:
        return super().stage()

    def unstage(self) -> List[object]:
        return super().unstage()

    def stop(self, *, success=False):
        self.controller.stop_all_axes()
        return super().stop(success=success)


if __name__ == "__main__":
    rtcontroller = RtFlomniController(
        socket_cls=SocketIO, socket_host="mpc2844.psi.ch", socket_port=2222
    )
    rtcontroller.on()
    rtcontroller.laser_tracker_on()



from ophyd import Component

from ophyd_devices.utils import bec_utils
from ophyd_devices.epics.devices.psi_delay_generator_base import (
    PSIDelayGeneratorBase,
    DDGCustomMixin,
    TriggerSource,
)

from bec_lib import bec_logger


logger = bec_logger.logger


class DelayGeneratorError(Exception):
    """Exception raised for errors."""


class DDGSetup(DDGCustomMixin):
    """
    Mixin class for DelayGenerator logic at cSAXS.

    At cSAXS, multiple DDGs were operated at the same time. There different behaviour is
        implemented in the ddg_config signals that are passed via the device config.
    """

    def initialize_default_parameter(self) -> None:
        """Method to initialize default parameters."""
        for ii, channel in enumerate(self.parent.all_channels):
            self.parent.set_channels("polarity", self.parent.polarity.get()[ii], [channel])

        self.parent.set_channels("amplitude", self.parent.amplitude.get())
        self.parent.set_channels("offset", self.parent.offset.get())
        # Setup reference
        self.parent.set_channels(
            "reference",
            0,
            [f"channel{pair}.ch1" for pair in self.parent.all_delay_pairs],
        )
        self.parent.set_channels(
            "reference",
            0,
            [f"channel{pair}.ch2" for pair in self.parent.all_delay_pairs],
        )
        self.parent.set_trigger(getattr(TriggerSource, self.parent.set_trigger_source.get()))
        # Set threshold level for ext. pulses
        self.parent.level.put(self.parent.thres_trig_level.get())

    def prepare_ddg(self) -> None:
        """
        Method to prepare scan logic of cSAXS

        Two scantypes are supported: "step" and "fly":
        - step: Scan is performed by stepping the motor and acquiring data at each step
        - fly: Scan is performed by moving the motor with a constant velocity and acquiring data

        Custom logic for different DDG behaviour during scans.

        - set_high_on_exposure      : If True, then TTL signal is high during
                                    the full exposure time of the scan (all frames).
                                    E.g. Keep shutter open for the full scan.
        - fixed_ttl_width           : fixed_ttl_width is a list of 5 values, one for each channel.
                                    If the value is 0, then the width of the TTL pulse is determined,
                                    no matter which parameters are passed from the scaninfo for exposure time
        - set_trigger_source        : Specifies the default trigger source for the DDG. For cSAXS, relevant ones
                                    were: SINGLE_SHOT, EXT_RISING_EDGE
        """
        self.parent.set_trigger(getattr(TriggerSource, self.parent.set_trigger_source.get()))
        # scantype "step"
        if self.parent.scaninfo.scan_type == "step":
            # High on exposure means that the signal
            if self.parent.set_high_on_exposure.get():
                # caluculate parameters
                num_burst_cycle = 1 + self.parent.additional_triggers.get()

                exp_time = (
                    self.parent.delta_width.get()
                    + self.parent.scaninfo.frames_per_trigger
                    * (self.parent.scaninfo.exp_time + self.parent.scaninfo.readout_time)
                )
                total_exposure = exp_time
                delay_burst = self.parent.delay_burst.get()

                # Set individual channel widths, if fixed_ttl_width and trigger_width are combined, this can be a common call too
                if not self.parent.trigger_width.get():
                    self.parent.set_channels("width", exp_time)
                else:
                    self.parent.set_channels("width", self.parent.trigger_width.get())
                for value, channel in zip(
                    self.parent.fixed_ttl_width.get(), self.parent.all_channels
                ):
                    logger.debug(f"Trying to set DDG {channel} to {value}")
                    if value != 0:
                        self.parent.set_channels("width", value, channels=[channel])
            else:
                # caluculate parameters
                exp_time = self.parent.delta_width.get() + self.parent.scaninfo.exp_time
                total_exposure = exp_time + self.parent.scaninfo.readout_time
                delay_burst = self.parent.delay_burst.get()
                num_burst_cycle = (
                    self.parent.scaninfo.frames_per_trigger + self.parent.additional_triggers.get()
                )

                # Set individual channel widths, if fixed_ttl_width and trigger_width are combined, this can be a common call too
                if not self.parent.trigger_width.get():
                    self.parent.set_channels("width", exp_time)
                else:
                    self.parent.set_channels("width", self.parent.trigger_width.get())
        # scantype "fly"
        elif self.parent.scaninfo.scan_type == "fly":
            if self.parent.set_high_on_exposure.get():
                # caluculate parameters
                exp_time = (
                    self.parent.delta_width.get()
                    + self.parent.scaninfo.exp_time * self.parent.scaninfo.num_points
                    + self.parent.scaninfo.readout_time * (self.parent.scaninfo.num_points - 1)
                )
                total_exposure = exp_time
                delay_burst = self.parent.delay_burst.get()
                num_burst_cycle = 1 + self.parent.additional_triggers.get()

                # Set individual channel widths, if fixed_ttl_width and trigger_width are combined, this can be a common call too
                if not self.parent.trigger_width.get():
                    self.parent.set_channels("width", exp_time)
                else:
                    self.parent.set_channels("width", self.parent.trigger_width.get())
                for value, channel in zip(
                    self.parent.fixed_ttl_width.get(), self.parent.all_channels
                ):
                    logger.debug(f"Trying to set DDG {channel} to {value}")
                    if value != 0:
                        self.parent.set_channels("width", value, channels=[channel])
            else:
                # caluculate parameters
                exp_time = self.parent.delta_width.get() + self.parent.scaninfo.exp_time
                total_exposure = exp_time + self.parent.scaninfo.readout_time
                delay_burst = self.parent.delay_burst.get()
                num_burst_cycle = (
                    self.parent.scaninfo.num_points + self.parent.additional_triggers.get()
                )

                # Set individual channel widths, if fixed_ttl_width and trigger_width are combined, this can be a common call too
                if not self.parent.trigger_width.get():
                    self.parent.set_channels("width", exp_time)
                else:
                    self.parent.set_channels("width", self.parent.trigger_width.get())

        else:
            raise Exception(f"Unknown scan type {self.parent.scaninfo.scan_type}")
        # Set common DDG parameters
        self.parent.burst_enable(num_burst_cycle, delay_burst, total_exposure, config="first")
        self.parent.set_channels("delay", 0.0)

    def on_trigger(self) -> None:
        """Method to be executed upon trigger"""
        if self.parent.source.read()[self.parent.source.name]["value"] == TriggerSource.SINGLE_SHOT:
            self.parent.trigger_shot.put(1)

    def check_scanID(self) -> None:
        """
        Method to check if scanID has changed.

        If yes, then it changes parent.stopped to True, which will stop further actions.
        """
        old_scanID = self.parent.scaninfo.scanID
        self.parent.scaninfo.load_scan_metadata()
        if self.parent.scaninfo.scanID != old_scanID:
            self.parent.stopped = True

    def finished(self) -> None:
        """Method checks if DDG finished acquisition"""

    def on_pre_scan(self) -> None:
        """
        Method called by pre_scan hook in parent class.

        Executes trigger if premove_trigger is Trus.
        """
        if self.parent.premove_trigger.get() is True:
            self.parent.trigger_shot.put(1)


class DelayGeneratorcSAXS(PSIDelayGeneratorBase):
    """
    DG645 delay generator at cSAXS (multiple can be in use depending on the setup)

    Default values for setting up DDG.
    Note: checks of set calues are not (only partially) included, check manual for details on possible settings.
    https://www.thinksrs.com/downloads/pdfs/manuals/DG645m.pdf

    - delay_burst               : (float >=0) Delay between trigger and first pulse in burst mode
    - delta_width               : (float >= 0) Add width to fast shutter signal to make sure its open during acquisition
    - additional_triggers       : (int) add additional triggers to burst mode (mcs card needs +1 triggers per line)
    - polarity                  : (list of 0/1) polarity for different channels
    - amplitude                 : (float) amplitude voltage of TTLs
    - offset                    : (float) offset for ampltitude
    - thres_trig_level          : (float) threshold of trigger amplitude

    Custom signals for logic in different DDGs during scans (for custom_prepare.prepare_ddg):

    - set_high_on_exposure      : (bool): if True, then TTL signal should go high during the full acquisition time of a scan.
    # TODO trigger_width and fixed_ttl could be combined into single list.
    - fixed_ttl_width           : (list of either 1 or 0), one for each channel.
    - trigger_width             : (float) if fixed_ttl_width is True, then the width of the TTL pulse is set to this value.
    - set_trigger_source        : (TriggerSource) specifies the default trigger source for the DDG.
    - premove_trigger           : (bool) if True, then a trigger should be executed before the scan starts (to be implemented in on_pre_scan).
    - set_high_on_stage         : (bool) if True, then TTL signal should go high already on stage.
    """

    custom_prepare_cls = DDGSetup

    delay_burst = Component(
        bec_utils.ConfigSignal,
        name="delay_burst",
        kind="config",
        config_storage_name="ddg_config",
    )

    delta_width = Component(
        bec_utils.ConfigSignal,
        name="delta_width",
        kind="config",
        config_storage_name="ddg_config",
    )

    additional_triggers = Component(
        bec_utils.ConfigSignal,
        name="additional_triggers",
        kind="config",
        config_storage_name="ddg_config",
    )

    polarity = Component(
        bec_utils.ConfigSignal,
        name="polarity",
        kind="config",
        config_storage_name="ddg_config",
    )

    fixed_ttl_width = Component(
        bec_utils.ConfigSignal,
        name="fixed_ttl_width",
        kind="config",
        config_storage_name="ddg_config",
    )

    amplitude = Component(
        bec_utils.ConfigSignal,
        name="amplitude",
        kind="config",
        config_storage_name="ddg_config",
    )

    offset = Component(
        bec_utils.ConfigSignal,
        name="offset",
        kind="config",
        config_storage_name="ddg_config",
    )

    thres_trig_level = Component(
        bec_utils.ConfigSignal,
        name="thres_trig_level",
        kind="config",
        config_storage_name="ddg_config",
    )

    set_high_on_exposure = Component(
        bec_utils.ConfigSignal,
        name="set_high_on_exposure",
        kind="config",
        config_storage_name="ddg_config",
    )

    set_high_on_stage = Component(
        bec_utils.ConfigSignal,
        name="set_high_on_stage",
        kind="config",
        config_storage_name="ddg_config",
    )

    set_trigger_source = Component(
        bec_utils.ConfigSignal,
        name="set_trigger_source",
        kind="config",
        config_storage_name="ddg_config",
    )

    trigger_width = Component(
        bec_utils.ConfigSignal,
        name="trigger_width",
        kind="config",
        config_storage_name="ddg_config",
    )
    premove_trigger = Component(
        bec_utils.ConfigSignal,
        name="premove_trigger",
        kind="config",
        config_storage_name="ddg_config",
    )

    def __init__(
        self,
        prefix="",
        *,
        name,
        kind=None,
        read_attrs=None,
        configuration_attrs=None,
        parent=None,
        device_manager=None,
        sim_mode=False,
        ddg_config=None,
        **kwargs,
    ):
        """
        Args:
            prefix (str, optional): Prefix of the device. Defaults to "".
            name (str): Name of the device.
            kind (str, optional): Kind of the device. Defaults to None.
            read_attrs (list, optional): List of attributes to read. Defaults to None.
            configuration_attrs (list, optional): List of attributes to configure. Defaults to None.
            parent (Device, optional): Parent device. Defaults to None.
            device_manager (DeviceManagerBase, optional): DeviceManagerBase object. Defaults to None.
            sim_mode (bool, optional): Simulation mode flag. Defaults to False.
            ddg_config (dict, optional): Dictionary of ddg_config signals. Defaults to None.

        """
        # Default values for ddg_config signals
        self.ddg_config = {
            # Setup default values
            f"{name}_delay_burst": 0,
            f"{name}_delta_width": 0,
            f"{name}_additional_triggers": 0,
            f"{name}_polarity": [1, 1, 1, 1, 1],
            f"{name}_amplitude": 4.5,
            f"{name}_offset": 0,
            f"{name}_thres_trig_level": 2.5,
            # Values for different behaviour during scans
            f"{name}_fixed_ttl_width": [0, 0, 0, 0, 0],
            f"{name}_trigger_width": None,
            f"{name}_set_high_on_exposure": False,
            f"{name}_set_high_on_stage": False,
            f"{name}_set_trigger_source": "SINGLE_SHOT",
            f"{name}_premove_trigger": False,
        }
        if ddg_config is not None:
            # pylint: disable=expression-not-assigned
            [self.ddg_config.update({f"{name}_{key}": value}) for key, value in ddg_config.items()]
        super().__init__(
            prefix=prefix,
            name=name,
            kind=kind,
            read_attrs=read_attrs,
            configuration_attrs=configuration_attrs,
            parent=parent,
            device_manager=device_manager,
            sim_mode=sim_mode,
            **kwargs,
        )


if __name__ == "__main__":
    # Start delay generator in simulation mode.
    # Note: To run, access to Epics must be available.
    dgen = DelayGeneratorcSAXS("delaygen:DG1:", name="dgen", sim_mode=True)
